# -*- coding: utf-8 -*-
#
#  meta.py
#  gazar
#
#  Created by Alan D Snow, 2017.
#  BSD 3-Clause
"""gazar.meta
    Module containing metadata for gazar
"""


def version():
    """
    Returns
    -------
    str: Version of gazar
    """
    return '0.0.3'
